#include <asm/hardware.h>


#ifdef CONFIG_EP93XX_GRAPHICS
static DECLARE_WAIT_QUEUE_HEAD(ep93xxfb_wait_in);

static irqreturn_t ep93xxfb_irq_handler(int i, void *blah)
{
	outl(0x00000000, EP93XX_GRAPHICS_ACCEL_BLOCKCTRL);
	wake_up(&ep93xxfb_wait_in);

	return IRQ_HANDLED;
}

static void ep93xxfb_wait(void)
{
	DECLARE_WAITQUEUE(wait, current);
	add_wait_queue(&ep93xxfb_wait_in, &wait);
	set_current_state(TASK_UNINTERRUPTIBLE);
	    
	while (inl(EP93XX_GRAPHICS_ACCEL_BLOCKCTRL) & 0x00000001 )
		if (!in_atomic())
			schedule();

	remove_wait_queue(&ep93xxfb_wait_in, &wait);
	set_current_state(TASK_RUNNING);
}

static int ep93xxfb_line(struct fb_info *info, struct ep93xx_line *line)
{
	unsigned long value = 0;
	long x, y, dx, dy, count, xinc, yinc, xval, yval, incr;
    
	if ((line->x1 > info->var.xres) ||
	    (line->x2 > info->var.xres) ||
	    (line->y1 > info->var.yres) ||
	    (line->y2 > info->var.yres))
		return -EFAULT;	
	x = line->x1;
	y = line->y1;
	dx = line->x2 - line->x1;
	dy = line->y2 - line->y1;

	if (!dx || !dy)
		return -EFAULT;

	if ((dx < 0) && (dy < 0)) {
		x = line->x2;
		y = line->y2;
		dx *= -1;
		dy *= -1;
	} else if ( dx < 0 && dy > 0 ) {
		dx *= -1;
		value = 0x000000A0;
	} else if( dy < 0 && dx > 0 ) {
		dy *= -1;
		value = 0x00000140;
	}
	
	if (line->flags & LINE_PRECISE) {
		count = isqrt(((dy * dy) + (dx * dx)) * 4096);
		xinc = (4095 * 64 * dx) / count;
		yinc = (4095 * 64 * dy) / count;
		xval = 2048;
		yval = 2048;
		count = 0;
		while (dx || dy) {
			incr = 0;
			xval -= xinc;
			if (xval <= 0) {
				xval += 4096;
				dx--;
				incr = 1;
			}
			yval -= yinc;
			if (yval <= 0) {
				yval += 4096;
				dy--;
				incr = 1;
			}
			count += incr;
		}
	} else {
		if ( dx == dy ) {
			xinc = 4095;
			yinc = 4095;
			count = dx;
			    
		} else if ( dx < dy ) {
			xinc = ( dx * 4095 ) / dy;
			yinc = 4095;
			count = dy;
				    
		} else {
			xinc = 4095;
			yinc = ( dy * 4095 ) / dx;
			count = dx;
		}
	}

	outl(0x08000800, EP93XX_GRAPHICS_ACCEL_LINEINIT);
	if (line->flags & LINE_PATTERN)
		outl(line->pattern, EP93XX_GRAPHICS_ACCEL_LINEPATTRN);
	else
		outl(0x000fffff, EP93XX_GRAPHICS_ACCEL_LINEPATTRN);
	outl(epinfo.fb_phys + ((y * epinfo.xres * epinfo.bpp) / 8) +
	     ((x * epinfo.bpp ) / 8 ), EP93XX_GRAPHICS_ACCEL_BLKDSTSTRT);

	outl(((x * epinfo.bpp) & 0x1F) |
	     ((((x + dx - 1) * epinfo.bpp) & 0x1F ) << 16),
	     EP93XX_GRAPHICS_ACCEL_DESTPIXELSTRT);
	outl((epinfo.xres * epinfo.bpp) / 32, EP93XX_GRAPHICS_ACCEL_DESTLINELENGTH);
	outl(line->fgcolor, EP93XX_GRAPHICS_ACCEL_BLOCKMASK);
	outl(line->bgcolor, EP93XX_GRAPHICS_ACCEL_BACKGROUND);
	outl((yinc << 16) | xinc, EP93XX_GRAPHICS_ACCEL_LINEINC);
	outl(count & 0xFFF, EP93XX_GRAPHICS_ACCEL_BLKDESTWIDTH);
	outl(0, EP93XX_GRAPHICS_ACCEL_BLKDESTHEIGHT);
	value |= (line->flags & LINE_BACKGROUND) ? 0x00004000 : 0;
	outl(epinfo.pixformat | value | 0x00000013, EP93XX_GRAPHICS_ACCEL_BLOCKCTRL);
	ep93xxfb_wait();

	return 0;
}

static void ep93xxfb_fillrect(struct fb_info *p, const struct fb_fillrect *fill)
{
	unsigned long blkdestwidth,tmp;

	if (!fill->width || !fill->height || 
	    (fill->dx >= p->var.xres) || 
	    (fill->dy >= p->var.yres) ||
	    ((fill->dx + fill->width - 1) >= p->var.xres) ||
	    ((fill->dy + fill->height - 1) >= p->var.yres))
		return;

	tmp =  (( fill->dx + fill->width ) * epinfo.bpp );
	blkdestwidth = tmp / 32;
	if(blkdestwidth > 0 && (tmp % 32 == 0))
		blkdestwidth--;
	blkdestwidth = blkdestwidth - (fill->dx * epinfo.bpp) / 32;
		       
	outl(fill->color, EP93XX_GRAPHICS_ACCEL_BLOCKMASK);
	outl(((fill->dx * epinfo.bpp) & 0x1F) |
	     ((((fill->dx + fill->width - 1) * epinfo.bpp ) & 0x1F) << 16),
	     EP93XX_GRAPHICS_ACCEL_DESTPIXELSTRT);
	outl(((epinfo.xres * epinfo.bpp) / 32), EP93XX_GRAPHICS_ACCEL_DESTLINELENGTH);
	outl(blkdestwidth, EP93XX_GRAPHICS_ACCEL_BLKDESTWIDTH );
	outl(fill->height - 1, EP93XX_GRAPHICS_ACCEL_BLKDESTHEIGHT);
	outl((epinfo.fb_phys + (fill->dy * epinfo.xres * epinfo.bpp ) / 8 +
	      (fill->dx * epinfo.bpp ) / 8 )
	     , EP93XX_GRAPHICS_ACCEL_BLKDSTSTRT);
	outl(epinfo.pixformat | 0x0000000B, EP93XX_GRAPHICS_ACCEL_BLOCKCTRL);
	ep93xxfb_wait();
}

void ep93xxfb_copyarea(struct fb_info *p, const struct fb_copyarea *area) 
{
	unsigned long startsx, stopsx, startdx, stopdx, startsy, startdy;
	unsigned long blksrcwidth, blkdestwidth;
	unsigned long tmp;
	unsigned long val = 0;

	if (!area->width || !area->width || 
	    (area->sx >= p->var.xres) || (area->sy >= p->var.yres) ||
	    (area->dx >= p->var.xres) || (area->dy >= p->var.yres) ||
	    ((area->dx + area->width - 1) >= p->var.xres) ||
	    ((area->dy + area->height - 1) >= p->var.yres))
		return;
    
	if ((area->sx == area->dx) && (area->sy == area->dy))
		return;

	if ((area->dy == area->sy) && (area->dx > area->sx) &&
	    (area->dx < (area->sx + area->width - 1))) {
		startdx = area->dx + area->width - 1;
		stopdx = area->dx;
		startsx = area->sx + area->width - 1;
		stopsx = area->sx;
		val |= 0x000000A0;
	} else {
		startdx = area->dx;
		stopdx = area->dx + area->width - 1;
		startsx = area->sx;
		stopsx = area->sx + area->width - 1;
	}

	if (area->dy <= area->sy) {
		startdy = area->dy;
		startsy = area->sy;
	} else {
		startdy = area->dy + area->height -1;
		startsy = area->sy + area->height -1;
		val |= 0x00000140;
	}
    
	tmp = (( area->sx + area->width ) * epinfo.bpp );
	blksrcwidth = tmp / 32;
	if (blksrcwidth > 0 && (tmp % 32 == 0))
		blksrcwidth--;
	blksrcwidth = blksrcwidth - (area->sx * epinfo.bpp) / 32;				           

	tmp = (( area->dx + area->width ) * epinfo.bpp );
	blkdestwidth = tmp / 32;
	if (blkdestwidth > 0 && (tmp % 32 == 0))
		blkdestwidth--;
	blkdestwidth = blkdestwidth - (area->dx * epinfo.bpp) / 32;

	outl(0x00000000, EP93XX_GRAPHICS_ACCEL_BLOCKCTRL);

	/*** src ***/
	outl((((startsx * epinfo.bpp) & 0x1F) |
	      (((stopsx * epinfo.bpp ) & 0x1F) << 16))
	     , EP93XX_GRAPHICS_ACCEL_SRCPIXELSTRT);	
	outl((epinfo.fb_phys + (startsy * epinfo.xres * epinfo.bpp ) / 8 +
	      (startsx * epinfo.bpp ) / 8 )
	     , EP93XX_GRAPHICS_ACCEL_BLKSRCSTRT);
	outl(((epinfo.xres * epinfo.bpp) / 32), EP93XX_GRAPHICS_ACCEL_SRCLINELENGTH);
	outl( blksrcwidth, EP93XX_GRAPHICS_ACCEL_BLKSRCWIDTH );
	    
	/*** dest ***/
	outl((((startdx * epinfo.bpp) & 0x1F) | 
	      (((stopdx * epinfo.bpp ) & 0x1F) << 16))
	     , EP93XX_GRAPHICS_ACCEL_DESTPIXELSTRT);
	outl((epinfo.fb_phys + (startdy * epinfo.xres * epinfo.bpp ) / 8 +
	      (startdx * epinfo.bpp ) / 8 )
	     , EP93XX_GRAPHICS_ACCEL_BLKDSTSTRT);
	outl(((epinfo.xres * epinfo.bpp) / 32), EP93XX_GRAPHICS_ACCEL_DESTLINELENGTH);
	outl(blkdestwidth, EP93XX_GRAPHICS_ACCEL_BLKDESTWIDTH);
	outl(area->height - 1 , EP93XX_GRAPHICS_ACCEL_BLKDESTHEIGHT);
	outl(epinfo.pixformat | val | 0x00000003, EP93XX_GRAPHICS_ACCEL_BLOCKCTRL);
	ep93xxfb_wait();
}
    
